
#include "shared.h"

byte bufb[0x200];               /* Buffer for plane B */
byte bufs[0x200];               /* Buffer for plane S */
byte bufi[0x200];               /* Internal buffer for 16-bit rendering */
byte cache[0x80000];            /* Expanded pattern cache */
byte lut[LUT_MAX][0x10000];     /* Look-up tables for merging */

word pixel[0x100];              /* 8 to 16-bit pixel conversion */

byte vram_dirty_list[0x800];    /* List of dirty patterns (by index) */
word vram_dirty_queue[0x800];   /* List of dirty patterns (by name) */
word vram_dirty_index=0;        /* Offset into queue */

int plane_a_enable;             /* Plane A, as clipped by plane W */
int plane_a_start;
int plane_a_end;

int plane_w_enable;             /* Plane W, as clipped by plane A */
int plane_w_start;
int plane_w_end;


static dword atex_table[] =     /* Attribute expansion table */
{
    0x00000000,       
    0x10101010,       
    0x20202020,       
    0x30303030,       
    0x40404040,       
    0x50505050,       
    0x60606060,       
    0x70707070,       
};

/*--------------------------------------------------------------------------*/
/* Line rendering functions                                                 */
/*--------------------------------------------------------------------------*/

void render_line(int line)
{
    int scrolla;
    int scrollb;

    /* Point to offset within bitmap */
    byte *lb = (bitmap.depth == 8) ? &bitmap.data[(line * bitmap.pitch)] : bufi;

    /* Draw line if display is turned on */
    if((vdp->reg[1] & 0x40) == 0x40)
    {
        scrolla = get_hscroll(0, line);
        scrollb = get_hscroll(1, line);
        if(vram_dirty_index) cache_tiles();
        window_clip(line);

        if(plane_a_enable) render_a(line, lb, scrolla);
        if(plane_w_enable) render_w(line, lb);

        if(plane_a_enable || plane_w_enable)
        {
            render_b(line, bufb, scrollb);
            merge(lb + BITMAP_MARGIN, bufb + BITMAP_MARGIN, vp_width, lut[LUT_BG]);
        }
        else
        {
            render_b(line, lb, scrollb);
        }

        memset(bufs + BITMAP_MARGIN, 0, vp_width);
        render_sprites(line, bufs, lut[LUT_OBJ]);
        merge(lb + BITMAP_MARGIN, bufs + BITMAP_MARGIN, vp_width, lut[(vdp->reg[12] & 8) ? LUT_BGOBJ_STE : LUT_BGOBJ]);
    }
    else /* Clear display (should be with overscan color...) */
    {
        memset(lb + BITMAP_MARGIN, 0, vp_width);
    }

    if(bitmap.depth == 16)
        remap(bufi + BITMAP_MARGIN, (word *)&bitmap.data[(line * bitmap.pitch) + (BITMAP_MARGIN << 1)], vp_width);
}


void render_a(int line, byte *lb, int scrolla)
{
    int column, attr;
    dword atex, *src, *dst;
    word *nt;
    word *vsc = (word *)&vdp->vsram[0];
    int vscroll = vsc[0];
    int end = plane_a_end + ((scrolla & 7) ? 1 : 0);
    
    if(vdp->reg[11] & 4) /* 2-cell vertical scroll mode */
    {
        for(column = plane_a_start; column < end; column += 1)
        {
            vscroll = vsc[(column & ~1)];
            vscroll = (line + (((vscroll >> 8) & 0xFF) | ((vscroll & 0xFF) << 8))) & pf_column_mask;
            nt = (word *)&vdp->vram[ntab + ((vscroll >> 3) << pf_shift)];

            attr = nt[(column + (scrolla >> 3)) & pf_row_mask];
            attr = ((attr >> 8) & 0xFF) | ((attr & 0xFF) << 8);
            atex = atex_table[(attr >> 13) & 7];

            src = (dword *)&cache[((attr & 0x1FFF) << 6) | ((vscroll & 7) << 3)];
            dst = (dword *)&lb[((column << 3) + BITMAP_MARGIN) - (scrolla & 7)];

            *dst++ = (*src++ | atex);
            *dst = (*src | atex);
        }
    }
    else
    {
        vscroll = (line + (((vscroll >> 8) & 0xFF) | ((vscroll & 0xFF) << 8))) & pf_column_mask;
        nt = (word *)&vdp->vram[ntab + ((vscroll >> 3) << pf_shift)];

        for(column = plane_a_start; column < end; column += 1)
        {
            attr = nt[(column + (scrolla >> 3)) & pf_row_mask];
            attr = ((attr >> 8) & 0xFF) | ((attr & 0xFF) << 8);
            atex = atex_table[(attr >> 13) & 7];

            src = (dword *)&cache[((attr & 0x1FFF) << 6) | ((vscroll & 7) << 3)];
            dst = (dword *)&lb[((column << 3) + BITMAP_MARGIN) - (scrolla & 7)];

            *dst++ = (*src++ | atex);
            *dst = (*src | atex);
        }
    }
}


void render_b(int line, byte *lb, int scrollb)
{
    int column, attr;
    dword atex, *src, *dst;
    word *nt;
    word *vsc = (word *)&vdp->vsram[0];
    int vscroll = vsc[1];
    int end = ((vdp->reg[12] & 1) ? 40 : 32) + ((scrollb & 7) ? 1 : 0);

    if(vdp->reg[11] & 4) /* 2-cell vertical scroll mode */
    {
        for(column = 0; column < end; column += 1)
        {
            vscroll = vsc[(column & ~1) | (1)];
            vscroll = (line + (((vscroll >> 8) & 0xFF) | ((vscroll & 0xFF) << 8))) & pf_column_mask;
            nt = (word *)&vdp->vram[ntbb + ((vscroll >> 3) << pf_shift)];

            attr = nt[(column + (scrollb >> 3)) & pf_row_mask];
            attr = ((attr >> 8) & 0xFF) | ((attr & 0xFF) << 8);
            atex = atex_table[(attr >> 13) & 7];

            src = (dword *)&cache[((attr & 0x1FFF) << 6) | ((vscroll & 7) << 3)];
            dst = (dword *)&lb[((column << 3) + BITMAP_MARGIN) - (scrollb & 7)];

            *dst++ = (*src++ | atex);
            *dst = (*src | atex);
        }
    }
    else
    {
        vscroll = (line + (((vscroll >> 8) & 0xFF) | ((vscroll & 0xFF) << 8))) & pf_column_mask;
        nt = (word *)&vdp->vram[ntbb + ((vscroll >> 3) << pf_shift)];

        for(column = 0; column < end; column += 1)
        {
            attr = nt[(column + (scrollb >> 3)) & pf_row_mask];
            attr = ((attr >> 8) & 0xFF) | ((attr & 0xFF) << 8);
            atex = atex_table[(attr >> 13) & 7];

            src = (dword *)&cache[((attr & 0x1FFF) << 6) | ((vscroll & 7) << 3)];
            dst = (dword *)&lb[((column << 3) + BITMAP_MARGIN) - (scrollb & 7)];

            *dst++ = (*src++ | atex);
            *dst = (*src | atex);
        }
    }
}


void render_w(int line, byte *lb)
{
    int column, attr;
    dword atex, *src, *dst;
    word *nt;

    nt = (word *)&vdp->vram[ntwb + ((line >> 3) << ((vdp->reg[12] & 1) ? 7 : 6))];
    for(column = plane_w_start; column < plane_w_end; column += 1)
    {
        attr = nt[column];
        attr = ((attr >> 8) & 0xFF) | ((attr & 0xFF) << 8);
        atex = atex_table[(attr >> 13) & 7];

        src = (dword *)&cache[((attr & 0x1FFF) << 6) | ((line & 7) << 3)];
        dst = (dword *)&lb[(column << 3) + BITMAP_MARGIN];

        *dst++ = (*src++ | atex);
        *dst = (*src | atex);
    }
}


void render_sprites(int line, byte *buf, byte *lut)
{
    int width;
    int height;
    int right_edge      = (vdp->reg[12] & 1) ? 448 : 384;
    int max             = (vdp->reg[12] & 1) ? 80 : 64;

    int count           = 0;
    int column          = 0;
    int link            = 0;
    int ypos            = 0;
    int xpos            = 0;
    int sprite_line     = 0;
    int nt_row          = 0;
    int v_row           = 0;
    int attr            = 0;

    byte *sat           = NULL;
    byte *src           = NULL;
    byte *dst           = NULL;

    for(count = 0; count < max; count += 1)
    {
        sat = &vdp->vram[satb + (link << 3)];
        ypos = (sat[0] << 8 | sat[1] ) & 0x1FF;
        xpos = (sat[6] << 8 | sat[7] ) & 0x1FF;
        width = (((sat[2] >> 2) & 3) << 3) + 8;
        height = ((sat[2] & 3) << 3) + 8;
        attr = (sat[4] << 8 | sat[5]);

/*      if(xpos == 0) break;  */
        if((count != 0) && (link == 0x00)) break;
        link = (sat[3] & 0x7F);

        if( ((line + 0x80) >= ypos) && ((line + 0x80) < (ypos + height)) && ((xpos + height) >= 0x80) && (xpos < right_edge) )
        {
            sprite_line = (line - (ypos - 0x80));
            nt_row = (sprite_line >> 3) & 3;
            v_row = (sprite_line & 7);

            if(attr & 0x1000)
                nt_row = (sat[2] & 3) - nt_row;

            for(column = 0; column < (width >> 3); column += 1)
            {                             
                int x, a, b, c;
                int ctrl = (attr >> 9) & 0x70;
                int name;
                int scol;

                scol = column;
                if(attr & 0x0800) scol = ((sat[2] >> 2) & 3) - column;

                name = (attr & 0x1800) | (((attr & 0x7FF) + (nt_row) + (scol * (height >> 3))) & 0x7FF);

                src = &cache[(name << 6) | (v_row << 3)];
                dst = &buf[BITMAP_MARGIN + (xpos - 0x80) + (column << 3)];

                for(x = 0; x < 8; x += 1)
                {
                    a = src[x];
                    if(a) {
                    b = dst[x];
                    c = lut[(b << 8) | (a | ctrl)];
                    dst[x] = c;
                    }
                }
            }
        }
    }
}

/*--------------------------------------------------------------------------*/
/* Helper functions                                                         */
/*--------------------------------------------------------------------------*/


void merge(byte *src, byte *dst, int length, byte *lut)
{
    do {
        *src++ = lut[((*dst++ << 8) | (*src))];
    } while (--length);
}


void remap(byte *src, word *dst, int length)
{
    do {
        *dst++ = pixel[(*src++)];
    } while (--length);
}


int get_hscroll(int plane, int line)
{
    word scroll = 0;
    word *hsc = (word *)&vdp->vram[hscb];

    switch(vdp->reg[11] & 3) {
        case 0: scroll = hsc[plane]; break; 
        case 1: scroll = hsc[((line &  7) << 1) | (plane)]; break; 
        case 2: scroll = hsc[((line & ~7) << 1) | (plane)]; break; 
        case 3: scroll = hsc[(line << 1) | (plane)]; break; 
    }

    scroll = 0x400 - (  (((scroll >> 8) & 0xFF) | ((scroll & 0xFF) << 8)) & 0x3FF);
    return (scroll);        
}


void cache_tiles(void)
{
    int x, y, c;
    int count;
    int name;

    for(count = 0; count < vram_dirty_index; count += 1)
    {
        name = vram_dirty_queue[count];
        vram_dirty_list[name] = 0;

        for(y = 0; y < 8; y += 1)
        {            
            dword line = *(dword *)&vdp->vram[(name << 5) | (y << 2)];

            for(x = 0; x < 8; x += 1)
            {
                c = (line >> ((x ^ 1) << 2)) & 0x0F;

                cache[ 0x00000 | (name << 6) | (y << 3) | (x) ] = c;
                cache[ 0x20000 | (name << 6) | (y << 3) | (x ^ 7) ] = c;
                cache[ 0x40000 | (name << 6) | ((y ^ 7) << 3) | (x) ] = c;
                cache[ 0x60000 | (name << 6) | ((y ^ 7) << 3) | (x ^ 7) ] = c;
            }            
        }
    }

    vram_dirty_index = 0;
}


void window_clip(int line)
{
    int vflip, vlen;
    int hflip, hlen;

    int rowStart, rowEnd, row = (line >> 3);

    /* Display size, in cells */
    int width = (vdp->reg[12] & 1) ? 40 : 32;
    int height = (vdp->reg[1] & 8) ? 30 : 28;

    /* Get clip vdp->register fields */
    hflip = (vdp->reg[17] >> 7) & 0x01;
    hlen  = ((vdp->reg[17] >> 0) & 0x1F) << 1;

    vflip = (vdp->reg[18] >> 7) & 0x01;
    vlen  = (vdp->reg[18] >> 0) & 0x1F;

    plane_a_enable  = 1;
    plane_a_start   = 0;
    plane_a_end     = width;

    plane_w_enable   = 0;
    plane_w_start    = 0;
    plane_w_end      = width;

    /* Window is disabled if clip registers are zero */
    if(((vdp->reg[17] & 0x9F) == 0x00) && ((vdp->reg[18] & 0x9F) == 0x00))
    {
        return;
    }

    /* Clip window row */
    rowStart = (vflip) ? vlen : 0;
    rowEnd   = (vflip) ? height : vlen;

    /* Clip current line to window */
    if((row >= rowStart) && (row < rowEnd))
    {
        plane_a_enable = 0;
        plane_w_enable  = 1;
        return;
    }

    /* Window isn't enabled on this row, check split point */
    if(hflip && (hlen >= width))
    {
        return;
    }

    /* Clip window column */
    if(hlen || hflip)
    {
        if(hflip)
        {
            plane_w_enable = 1;
            plane_w_start  = hlen;
            plane_w_end    = width;

            plane_a_enable = 1;
            plane_a_start  = 0;
            plane_a_end    = hlen;
        }
        else
        {
            plane_w_enable = 1;
            plane_w_start  = 0;
            plane_w_end    = hlen;

            plane_a_enable  = 1;
            plane_a_start   = hlen;
            plane_a_end     = width;
        }
    }
}


/*--------------------------------------------------------------------------*/
/* Table generation                                                         */
/*--------------------------------------------------------------------------*/


void make_table(void)
{
    int bx, ax;

    for(bx = 0; bx < 0x100; bx += 1)
    for(ax = 0; ax < 0x100; ax += 1)
    {
        word index = (bx << 8) | (ax);
        lut[ LUT_BG        ][index] = make_lut_bg(bx, ax);
        lut[ LUT_OBJ       ][index] = make_lut_obj(bx, ax);
        lut[ LUT_BGOBJ     ][index] = make_lut_bgobj(bx, ax);
        lut[ LUT_BGOBJ_STE ][index] = make_lut_bgobj_ste(bx, ax);
    }
}

int make_lut_bg(int bx, int ax)
{
    int bf, bp, b;
    int af, ap, a;
    int c;

    bf = (bx & 0x7F);
    bp = (bx >> 6) & 1;
    b  = (bx & 0x0F);
    
    af = (ax & 0x7F);   
    ap = (ax >> 6) & 1;
    a  = (ax & 0x0F);

    c = (ap ? (a ? af : bf) : \
        (bp ? (b ? bf : af) : \
        (     (a ? af : bf) )));

    if((c & 0x0F) == 0x00) c = 0x00;

    return (c);
}


int make_lut_obj(int bx, int sx)
{
    int bf, bs;
    int sf;
    int c;

    bf = (bx & 0x7F);   
    bs = (bx >> 7) & 1; 
    sf = (sx & 0x7F);
  
    c = (bs) ? bf : sf;

    return (c | 0x80);

}


int make_lut_bgobj(int bx, int ax)
{
    int bf, bp, b;
    int af, ap, a;
    int c;

    c = ax;
    ax = bx;
    bx = c;

    bf = (bx & 0x3F);
    bp = (bx >> 6) & 1;
    b  = (bx & 0x0F);
    
    af = (ax & 0x3F) | 0x40;
    ap = (ax >> 6) & 1;
    a  = (ax & 0x0F);

    c = (ap ? (a ? af : bf) : \
        (bp ? (b ? bf : af) : \
        (     (a ? af : bf) )));

    return (c);
}


int make_lut_bgobj_ste(int sx, int bx)
{
    int bf, bp, b;
    int sf, sp, s;
    int c, i = 0;

    bf = (bx & 0x3F);
    bp = (bx >> 6) & 1;
    b  = (bx & 0x0F);
    
    sf = (sx & 0x3F) | 0x40;
    sp = (sx >> 6) & 1;
    s  = (sx & 0x0F);

    if((sx & 0x3E) == 0x3E)  i = (sx & 1) ? 0x80 : 0xC0;

    if(i) {
        if(!sp && bp) return (bf);
        return (bf | i);
    }
    else
    {
        c = (sp ? (s ? sf : bf) : \
            (bp ? (b ? bf : sf) : \
            (     (s ? sf : bf) )));
    }

    return (c);
}
