
#ifndef _SYSTEM_H_
#define _SYSTEM_H_

/* YM2612 clock */
#define FM_CLOCK        (7670442)   /* 7.67 MHz (Maybe this should be 10MHz) */

/* PSG clock */
#define PSG_CLOCK       (3579545)   /* 3.58 MHz */

/* Input bitmasks player 1 & 2 inputs */
#define INPUT_UP        (0x00000001)
#define INPUT_DOWN      (0x00000002)
#define INPUT_LEFT      (0x00000004)
#define INPUT_RIGHT     (0x00000008)
#define INPUT_A         (0x00000010)
#define INPUT_B         (0x00000020)
#define INPUT_C         (0x00000040)

/* Input bitmasks for control panel */
#define INPUT_1PSTART   (0x00000001)
#define INPUT_2PSTART   (0x00000002)
#define INPUT_1PCOIN    (0x00000004)
#define INPUT_2PCOIN    (0x00000008)
#define INPUT_SERVICE   (0x00000010)
#define INPUT_TEST      (0x00000020)

/* Display */
typedef struct
{
    unsigned char *data;
    int width;
    int height;
    int depth;
    int pitch;
    struct
    {
        byte color[0x100][4];
        byte dirty[0x100];
        byte update;
    }palette;
} t_bitmap;

/* Sound */
typedef struct
{
    int enabled;
    int buffer_size;
    int sample_rate;
    signed short *fm_buffer[2];
    signed short *psg_buffer;
    signed short *pcm_buffer;
    signed short *buffer[2];
} t_snd;

/* Input */
typedef struct
{
    int pad[2];
    int system;
} t_input;

/* File path structure */
/* Not really sure how to handle this in a suitable manner, but having
   to allocate strings gets too complex rather quickly */
typedef struct
{
    char rom[0x100];
    char zip[0x100];
    char dip[0x100];
    char snap[0x100];
} t_path;

/* Global data */
extern t_bitmap bitmap;
extern t_input input;
extern t_snd snd;
extern t_path dirs;

/* Function prototypes */
int system_init(int sample_rate);
int audio_init(int sample_rate);
void system_reset(void);
void system_shutdown(void);

#endif /* _SYSTEM_H_ */
