/*
    More information:
    http://cgfm2.emuviews.com/genvdp.txt
*/

#include "shared.h"
#include "vcnt.h"
#include "math.h"


/* Add a pattern name to the queue if it hasn't been flagged as dirty */
#define MARK_DIRTY                                  \
{                                                   \
    int name = (vdp->addr >> 5) & 0x7FF;            \
    if(vram_dirty_list[name] == 0)                  \
    {                                               \
        vram_dirty_list[name] = 1;                  \
        vram_dirty_queue[vram_dirty_index] = name;  \
        vram_dirty_index += 1;                      \
    }                                               \
}                                                   \

/* VDP context */
static t_vdp vdp_context;
t_vdp *vdp = &vdp_context;

/* Base address of name, sprite, and scroll tables */
word ntab, ntbb, ntwb, satb, hscb;

/* Playfield dimensions */
int pf_shift;
int pf_row_mask;
int pf_column_mask;

/* Physical display dimensions */
int vp_changed;
int vp_width;
int vp_height;

/*--------------------------------------------------------------------------*/
/* Memory handlers                                                          */
/*--------------------------------------------------------------------------*/

void vdp_w_8(int address, int data)
{
    switch(address)
    {
        case 0xC00000: /* DATA */
        case 0xC00001:
        case 0xC00002:
        case 0xC00003:
            vdp_data_w(data << 8 | data);
            break;

        case 0xC00004: /* CTRL */
        case 0xC00006:
        case 0xC00005:
        case 0xC00007:
            vdp_ctrl_w(data << 8 | data);
            break;

        case 0xC00011:
        case 0xC00013:
        case 0xC00015:
        case 0xC00017:
            if(snd.enabled) SN76496Write(0, data);
            break;
    }
}

void vdp_w_16(int address, int data)
{
    switch(address)
    {
        case 0xC00000: /* DATA */
        case 0xC00002:
            vdp_data_w(data);
            break;

        case 0xC00004: /* CTRL */
        case 0xC00006:
            vdp_ctrl_w(data);
            break;
    }
}

int vdp_r_8(int address)
{
    switch(address)
    {
        case 0xC00000: /* DATA (MSB) */
        case 0xC00002:
            return ((vdp_data_r() >> 8) & 0xFF);

        case 0xC00001: /* DATA (LSB) */
        case 0xC00003:
            return ((vdp_data_r() >> 0) & 0xFF);

        case 0xC00004: /* CTRL (MSB) */
        case 0xC00006:
            return ((vdp_ctrl_r() >> 8) & 0xFF);

        case 0xC00005: /* CTRL (LSB) */
        case 0xC00007:
            return ((vdp_ctrl_r() >> 0) & 0xFF);

        case 0xC00008: /* HV COUNTER (MSB) */
        case 0xC0000A:
        case 0xC0000C:
        case 0xC0000E:
            return ((vdp_hv_counter() >> 8) & 0xFF);

        case 0xC00009: /* HV COUNTER (LSB) */
        case 0xC0000B:
        case 0xC0000D:
        case 0xC0000F:
            return ((vdp_hv_counter() >> 0) & 0xFF);
    }

    return (0xFF);
}

int vdp_r_16(int address)
{
    switch(address)
    {
        case 0xC00000: /* DATA */
        case 0xC00002:
            return (vdp_data_r());

        case 0xC00004: /* CTRL */
        case 0xC00006:
            return (vdp_ctrl_r());

        case 0xC00008: /* HV COUNTER */
        case 0xC0000A:
        case 0xC0000C:
        case 0xC0000E:
            return (vdp_hv_counter());
    }

    return (0xFFFF);
}

/*--------------------------------------------------------------------------*/
/* Data, control, and HV counter                                            */
/*--------------------------------------------------------------------------*/

int vdp_data_r(void)
{
    word temp = 0xFFFF;
    vdp->pending = 0;

    switch(vdp->code & 0x0F)
    {
        case VRAM_READ:
            temp = ( vdp->vram[(vdp->addr & 0xFFFE) | (0)] << 8 | vdp->vram[(vdp->addr & 0xFFFE) | (1)] );
            break;

        case CRAM_READ:
            temp = ( vdp->cram[(vdp->addr & 0x7E) | (0)] << 8 | vdp->cram[(vdp->addr & 0x7E) | (1)] );
            break;

        case VSRAM_READ:
            temp = ( vdp->vsram[(vdp->addr & 0x7E) | (0)] << 8 | vdp->vsram[(vdp->addr & 0x7E) | (1)] );
            break;

        default:
            error("Bad code for data read (A:%04X C:%02X)\n", vdp->addr, vdp->code);
            break;
    }

    vdp->addr += vdp->reg[15];
    return (temp);
}


void vdp_data_w(int data)
{
    vdp->pending = 0;

    if(vdp->fill == 1)
    {
        vdp_dma_fill(data);
        vdp->fill = 0;
        return;
    }

    switch(vdp->code & 0x0F)
    {
        case VRAM_WRITE:
            vdp->vram[((vdp->addr & 0xFFFE) | (0)) ^ (vdp->addr & 1)] = (data >> 8) & 0xFF;
            vdp->vram[((vdp->addr & 0xFFFE) | (1)) ^ (vdp->addr & 1)] = (data >> 0) & 0xFF;
            MARK_DIRTY;
            break;

        case CRAM_WRITE:
            vdp->cram[(vdp->addr & 0x7E) | (0)] = (data >> 8) & 0xFF;
            vdp->cram[(vdp->addr & 0x7E) | (1)] = (data >> 0) & 0xFF;
            break;

        case VSRAM_WRITE:
            vdp->vsram[(vdp->addr & 0x7E) | (0)] = (data >> 8) & 0xFF;
            vdp->vsram[(vdp->addr & 0x7E) | (1)] = (data >> 0) & 0xFF;
            break;

        default:
            if((vdp->code & 0x0F) != 0x00) error("Bad code for data write (A:%04X C:%02X D:%04X)\n", vdp->addr, vdp->code, data);
            break;
    }

    vdp->addr += vdp->reg[15];
}


int vdp_ctrl_r(void)
{
    vdp->pending = 0;
    return (vdp->status);
}


void vdp_ctrl_w(int data)
{
    if(vdp->pending == 0)
    {
        if((data & 0xC000) == 0x8000)
        {
            vdp_reg_w(data);
            vdp->code = 0;
        }
        else
        {
            vdp->pending = 1;

            vdp->addr = ((vdp->addr & 0xC000) | (data & 0x3FFF)) & 0xFFFF;
            vdp->code = ((vdp->code & 0x3C) | ((data >> 14) & 0x03)) & 0x3F;
        }
    }
    else
    {
        vdp->pending = 0;

        vdp->addr = ((vdp->addr & 0x3FFF) | ((data << 14) & 0xC000)) & 0xFFFF;
        vdp->code = ((vdp->code & 0x03) | ((data >> 2) & 0x3C)) & 0x3F;

        if((vdp->code & 0x20) && (vdp->reg[1] & 0x10))
        {
            switch(vdp->reg[23] & 0xC0)
            {
                case 0x00: 
                case 0x40: /* VDP DMA */
                    vdp_dma_68k();
                    break;

                case 0x80: /* VRAM fill */
                    vdp->fill = 1;
                    break;

                case 0xC0: /* VRAM copy */
                    vdp_dma_copy();
                    break;
            }
        }
    }
}


int vdp_hv_counter(void)
{
    word hcounter, vcounter;
    int cycles;
    int ret;

    vcounter = vcnt[(vdp->curline & 0x1FF)];
    cycles = (m68k_get_elapsed_cycles() % 1024);

    if(CELLMODE)
    {
//      float step = (509.0 / 256.0);
//      hcounter = (float)cycles / step;
        hcounter = (cycles >> 1);
    }
    else
    {
        hcounter = (cycles >> 2);
    }

    ret = (vcounter & 0xFF) << 8 | (hcounter & 0xFF);
    return (ret);
}

/*--------------------------------------------------------------------------*/
/* DMA functions                                                            */
/*--------------------------------------------------------------------------*/

void vdp_dma_fill(int data)
{
    int count;
    int length = (vdp->reg[19] | vdp->reg[20] << 8) & 0xFFFF;
    if(!length) length = 0xffff;

    if((vdp->code & 0x0F) != VRAM_WRITE)
        error("Bad vdp->code for VRAM fill (C:%02X A:%04X L:%04X %08X)\n", vdp->code, vdp->addr, length, m68k_readpc());

    vdp->vram[(vdp->addr & 0xFFFF)] = (data >> 0) & 0xFF;
    for(count = 0; count <= length; count += 1)
    {
        vdp->vram[(vdp->addr & 0xFFFF) ^ 1] = (data >> 8) & 0xFF;
        MARK_DIRTY;
        vdp->addr += vdp->reg[15];
    }
}


void vdp_dma_copy(void)
{
    int count;
    int length = (vdp->reg[19] | vdp->reg[20] << 8) & 0xFFFF;
    int source = (vdp->reg[21] | vdp->reg[22] << 8) & 0xFFFF;
    if(!length) length = 0xFFFF;

    if((vdp->code & 0x0F) != VRAM_READ)
        error("Bad vdp->code for VRAM copy (S:%04X D:%04X L:%04X C:%02X I:%02X %08X)\n", source, vdp->addr, length, vdp->code, vdp->reg[15], m68k_readpc());

    for(count = 0; count < length; count += 1)
    {
        byte temp = vdp->vram[source];
        vdp->vram[vdp->addr] = temp;

        source += 1;
        vdp->addr += vdp->reg[15];

        MARK_DIRTY;
    }
}


void vdp_dma_68k(void)
{
    int count;
    int length = (vdp->reg[19] | vdp->reg[20] << 8) & 0xFFFF;
    int source = (vdp->reg[21] << 1 | vdp->reg[22] << 9 | (vdp->reg[23] & 0x7F) << 17) & 0xFFFFFE;
    if(!length) length = 0xFFFF;

    for(count = 0; count < length; count += 1)
    {
        vdp_data_w(m68k_fetch(source));
        source += 2;
    }
}

/*--------------------------------------------------------------------------*/
/* Miscellaneous functions                                                  */
/*--------------------------------------------------------------------------*/

void vdp_reg_w(int data)
{
    int r = (data >> 8) & 0x1F;
    int d = (data >> 0) & 0xFF;

    switch(r)
    {
        case 0x01:
            if((d & 8) && (vp_height != 240))
            {
                vp_changed = 1;
                vp_height = 240;
            }
            if(!(d & 8) && (vp_height != 224))
            {
                vp_changed = 1;
                vp_height = 224;
            }
            break;

        case 0x0C:
            if((d & 1) && (vp_width != 320))
            {
                vp_changed = 1;
                vp_width = 320;
            }
            if(!(d & 1) && (vp_width != 256))
            {
                vp_changed = 1;
                vp_width = 256;
            }
            break;

        case 16:
            {
                static int pf_size_table[4] = {32, 64, 128, 128};
                static int pf_shift_table[4] = {6, 7, 8, 8};
                pf_shift = pf_shift_table[(d >> 0) & 3];
                pf_row_mask = (pf_size_table[(d >> 0) & 3] - 1);
                pf_column_mask = (pf_size_table[(d >> 4) & 3] << 3) - 1;
            }
            break;

        default:
            break;
    }

    vdp->reg[r] = d;
    table_sync();
}


void table_sync(void)
{
    ntab = (vdp->reg[2] & 0x38) << 10;
    ntwb = (vdp->reg[3] & (CELLMODE ? 0x3E : 0x3F)) << 10;
    ntbb = (vdp->reg[4] & 0x07) << 13;
    satb = (vdp->reg[5] & (CELLMODE ? 0x7E : 0x7F)) << 9;
    hscb = (vdp->reg[13] & 0x3F) << 10;
}

void vdp_reset(void)
{
    int i;
    static byte init[24] =
    {
        0x04, 0x44, 0x30, 0x3C,
        0x07, 0x6C, 0x00, 0x00,
        0x00, 0x00, 0xFF, 0x00,
        0x00, 0x37, 0x00, 0x02,
        0x01, 0x00, 0x00, 0xFF,
        0xFF, 0x00, 0x00, 0x80,
    };

    /* Clear the VDP context */
    memset(vdp, 0, sizeof(t_vdp));

    /* Initialize all the related variables */
    for(i = 0; i < 24; i += 1)
    {
        vdp_reg_w(0x8000 | (i << 8) | init[i]);
    }
}




