
#ifndef _VDP_H_
#define _VDP_H_

/* Display dimensions */
#define DISPLAY_WIDTH       ((vdp->reg[1] & 1) ? 240 : 224)
#define DISPLAY_HEIGHT      ((vdp->reg[12] & 1) ? 320 : 256)
#define CELLMODE            (vdp->reg[12] & 1)

/* Status flag bits */
#define STATUS_D15          (0x8000)
#define STATUS_D14          (0x4000)
#define STATUS_D13          (0x2000)
#define STATUS_D12          (0x1000)
#define STATUS_D11          (0x0800)
#define STATUS_D10          (0x0400)
#define STATUS_FIFOEMPTY    (0x0200)
#define STATUS_FIFOFULL     (0x0100)
#define STATUS_VINT         (0x0080)
#define STATUS_SPROVER      (0x0040)
#define STATUS_SPRCOL       (0x0020)
#define STATUS_ODDFRAME     (0x0010)
#define STATUS_VBLANK       (0x0008)
#define STATUS_HBLANK       (0x0004)
#define STATUS_DMABUSY      (0x0002)
#define STATUS_PALMODE      (0x0001)

/* Code register settings */
#define VRAM_READ           (0x00)
#define VRAM_WRITE          (0x01)
#define CRAM_WRITE          (0x03)
#define VSRAM_READ          (0x04)
#define VSRAM_WRITE         (0x05)
#define CRAM_READ           (0x08)

typedef struct
{
    byte vram[0x10000];
    byte cram[0x80];
    byte vsram[0x80];
    byte reg[0x20];
    byte pending;
    byte code;
    word addr;
    word status;
    byte fill;
    int curline;
    int counter;
}t_vdp;

extern t_vdp *vdp;

/* Global data */
extern word ntab;
extern word ntbb;
extern word ntwb;
extern word satb;
extern word hscb;

extern int pf_shift;
extern int pf_row_mask;       
extern int pf_column_mask;

extern int vp_changed;
extern int vp_width;
extern int vp_height;

/* Function prototypes */
void vdp_w_8(int address, int data);
void vdp_w_16(int address, int data);
int vdp_r_8(int address);
int vdp_r_16(int address);
void vdp_reset(void);
int vdp_data_r(void);
void vdp_data_w(int data);
int vdp_ctrl_r(void);
void vdp_ctrl_w(int data);
int vdp_hv_counter(void);
void vdp_dma_fill(int data);
void vdp_dma_copy(void);
void vdp_dma_68k(void);
void vdp_reg_w(int data);
void table_sync(void);

#endif /* _VDP_H_ */
