
#include "shared.h"

/* The following data is only valid for the current driver loaded, and
   will be initialized in the load and free driver functions */

byte dip_used;          /* 1= DIP switches were modified and will be saved */
byte dip_bank;          /* Editing DIP switch #1 or #2 */
byte dip_switch[2];     /* DIP switch values for both banks */
t_driver *game_driver;  /* Pointer to currently loaded game driver */


/*--------------------------------------------------------------------------*/
/* Bloxeed (System C)                                                       */
/*--------------------------------------------------------------------------*/

t_rominfo bloxeedc_rom[] =
{
    {"12908.bin", rom, 0x020000, 0x000000, ROM_LOAD_EVEN },
    {"12907.bin", rom, 0x020000, 0x000000, ROM_LOAD_ODD },
    {"12993.bin", rom, 0x020000, 0x040000, ROM_LOAD_EVEN },
    {"12992.bin", rom, 0x020000, 0x040000, ROM_LOAD_ODD },
    { NULL }
};


t_driver bloxeedc_driver =
{
    "bloxeedc\0",
    "Bloxeed\0",
    bloxeedc_rom,
    bloxeedc_init,
};


/*--------------------------------------------------------------------------*/
/* Columns 2 (System C)                                                     */
/*--------------------------------------------------------------------------*/

t_rominfo columns2_rom[] =
{
    {"epr13361.rom", rom, 0x020000, 0x000000, ROM_LOAD_EVEN },
    {"epr13360.rom", rom, 0x020000, 0x000000, ROM_LOAD_ODD },
    { NULL }
};

t_driver columns2_driver =
{
    "columns2\0",
    "Columns II - The Voyage Through Time (Japan)\0",
    columns2_rom,
    columns2_init,
};


/*--------------------------------------------------------------------------*/
/* Puyo Puyo                                                                */
/*--------------------------------------------------------------------------*/

t_rominfo puyopuyo_rom[] =
{
    {"epr15036", rom, 0x020000, 0x000000, ROM_LOAD_EVEN },
	{"epr15035", rom, 0x020000, 0x000000, ROM_LOAD_ODD },
    {"epr15038", rom, 0x020000, 0x100000, ROM_LOAD_EVEN },
    {"epr15037", rom, 0x020000, 0x100000, ROM_LOAD_ODD },
    {"epr15034", pcm, 0x020000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver puyopuyo_driver =
{
    "puyopuyo\0",
    "Puyo Puyo (Program Revision ?, Japan)\0",
    puyopuyo_rom,
    puyopuyo_init,
};

/*--------------------------------------------------------------------------*/
/* Puyo Puyo (Program Revision A)                                           */
/*--------------------------------------------------------------------------*/

t_rominfo puyopuya_rom[] =
{
    {"ep15036a.32", rom, 0x020000, 0x000000, ROM_LOAD_EVEN },
	{"ep15035a.31", rom, 0x020000, 0x000000, ROM_LOAD_ODD },
    {"epr15038", rom, 0x020000, 0x100000, ROM_LOAD_EVEN },
	{"epr15037", rom, 0x020000, 0x100000, ROM_LOAD_ODD },
    {"epr15034", pcm, 0x020000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver puyopuya_driver =
{
    "puyopuya\0",
    "Puyo Puyo (Program Revsion A, Japan)\0",
    puyopuya_rom,
    puyopuya_init,
};

/*--------------------------------------------------------------------------*/
/* PotoPoto                                                                 */
/*--------------------------------------------------------------------------*/

t_rominfo potopoto_rom[] =
{
    {"epr16662", rom, 0x040000, 0x000000, ROM_LOAD_EVEN },
    {"epr16661", rom, 0x040000, 0x000000, ROM_LOAD_ODD },
    {"epr16660", pcm, 0x040000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver potopoto_driver =
{
    "potopoto\0",
    "PotoPoto (Japan)\0",
    potopoto_rom,
    potopoto_init,
};



/*--------------------------------------------------------------------------*/
/* Columns                                                                  */
/*--------------------------------------------------------------------------*/

t_rominfo columns_rom[] =
{
    {"epr13112.32", rom, 0x020000, 0x000000, ROM_LOAD_EVEN },
    {"epr13111.31", rom, 0x020000, 0x000000, ROM_LOAD_ODD },
    { NULL }
};

t_driver columns_driver =
{
    "columns\0",
    "Columns (Japan)\0",
    columns_rom,
    columns_init,
};


/*--------------------------------------------------------------------------*/
/* Stack Columns                                                            */
/*--------------------------------------------------------------------------*/

t_rominfo stkclmns_rom[] =
{
    {"epr16795.32", rom, 0x080000, 0x000000, ROM_LOAD_EVEN },
    {"epr16794.31", rom, 0x080000, 0x000000, ROM_LOAD_ODD },
    {"mpr16797.34", rom, 0x080000, 0x100000, ROM_LOAD_EVEN },
    {"mpr16796.33", rom, 0x080000, 0x100000, ROM_LOAD_ODD }, 
    { NULL }
};

t_driver stkclmns_driver =
{
    "stkclmns\0",
    "Stack Columns (Japan)\0",
    stkclmns_rom,
    stkclmns_init,
};

/*--------------------------------------------------------------------------*/
/* Zunzunkyou No Yabou                                                      */
/*--------------------------------------------------------------------------*/

t_rominfo zunkyou_rom[] =
{
    {"epr16812.32", rom, 0x080000, 0x000000, ROM_LOAD_EVEN },
    {"epr16811.31", rom, 0x080000, 0x000000, ROM_LOAD_ODD },
    {"epr16814.34", rom, 0x080000, 0x100000, ROM_LOAD_EVEN },
    {"epr16813.33", rom, 0x080000, 0x100000, ROM_LOAD_ODD },
    {"epr16810.4",  pcm, 0x080000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver zunkyou_driver =
{
    "zunkyou\0",
    "Zunzunkyou No Yabou (Japan)\0",
    zunkyou_rom,
    zunkyou_init,
};


/*--------------------------------------------------------------------------*/
/* Borench                                                                  */
/*--------------------------------------------------------------------------*/

t_rominfo borench_rom[] =
{
    {"ic32.bin", rom, 0x040000, 0x000000, ROM_LOAD_EVEN },
    {"ic31.bin", rom, 0x040000, 0x000000, ROM_LOAD_ODD },
    {"ic4.bin",  pcm, 0x020000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver borench_driver =
{
    "borench\0",
    "Borench\0",
    borench_rom,
    borench_init,
};

/*--------------------------------------------------------------------------*/
/* Thunder Force AC (US/Europe Bootleg)                                     */
/*--------------------------------------------------------------------------*/

t_rominfo tfacbl_rom[] =
{
    {"4.bin", rom, 0x040000, 0x000000, ROM_LOAD_EVEN },  /* Unique to Bootleg */
    {"3.bin", rom, 0x040000, 0x000000, ROM_LOAD_ODD },   /* Unique to Bootleg */
    {"ic34.bin", rom, 0x040000, 0x100000, ROM_LOAD_EVEN },
    {"ic33.bin", rom, 0x040000, 0x100000, ROM_LOAD_ODD },
    {"1.bin", pcm, 0x020000, 0x000000, ROM_LOAD },  /* Unique to Bootleg */
    {"2.bin", pcm, 0x020000, 0x020000, ROM_LOAD },  /* Unique to Bootleg */
    { NULL }
};

t_driver tfacbl_driver =
{
    "tfacbl\0",
    "Thunder Force AC (US/Europe, Bootleg)\0",
    tfacbl_rom,
    tfacbl_init,
};

/*--------------------------------------------------------------------------*/
/* Thunder Force AC (US)                                               */
/*--------------------------------------------------------------------------*/

t_rominfo tforceac_rom[] =
{
    {"ic32.bin", rom, 0x040000, 0x000000, ROM_LOAD_EVEN },
    {"ic31.bin", rom, 0x040000, 0x000000, ROM_LOAD_ODD },
    {"ic34.bin", rom, 0x040000, 0x100000, ROM_LOAD_EVEN },
    {"ic33.bin", rom, 0x040000, 0x100000, ROM_LOAD_ODD },
    {"ic4.bin", pcm, 0x040000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver tforceac_driver =
{
    "tforceac\0",
    "Thunder Force AC (US/Europe)\0",
    tforceac_rom,
    tforceac_init,
};

/*--------------------------------------------------------------------------*/
/* Thunder Force AC (Japan)                                                 */
/*--------------------------------------------------------------------------*/

t_rominfo tfacjp_rom[] =
{
    {"Epr13657.32", rom, 0x040000, 0x000000, ROM_LOAD_EVEN }, /* Unique to Jpn */
    {"Epr13656.31", rom, 0x040000, 0x000000, ROM_LOAD_ODD },  /* Unique to Jpn */
    {"ic34.bin", rom, 0x040000, 0x100000, ROM_LOAD_EVEN },
    {"ic33.bin", rom, 0x040000, 0x100000, ROM_LOAD_ODD },
    {"ic4.bin", pcm, 0x040000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver tfacjp_driver =
{
    "tfacjp\0",
    "Thunder Force AC (Japan)\0",
    tfacjp_rom,
    tfacjp_init,
};

/*--------------------------------------------------------------------------*/
/* Puyo Puyo 2                                                              */
/*--------------------------------------------------------------------------*/

t_rominfo puyopuy2_rom[] =
{
    {"pp2.32", rom, 0x080000, 0x000000, ROM_LOAD_EVEN },
    {"pp2.31", rom, 0x080000, 0x000000, ROM_LOAD_ODD },
    {"pp2.4", pcm, 0x080000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver puyopuy2_driver =
{
    "puyopuy2\0",
    "Puyo Puyo 2 (Expert?) (Japan)\0",
    puyopuy2_rom,
    puyopuy2_init,
};

/*--------------------------------------------------------------------------*/
/*  Tant-R (Puzzle & Action) (Japan)                                        */
/*--------------------------------------------------------------------------*/

t_rominfo tantr_rom[] =
{
    {"epr15614.32", rom, 0x080000, 0x000000, ROM_LOAD_EVEN },
    {"epr15613.31", rom, 0x080000, 0x000000, ROM_LOAD_ODD },
    {"mpr15616.34", rom, 0x080000, 0x100000, ROM_LOAD_EVEN },
    {"mpr15615.33", rom, 0x080000, 0x100000, ROM_LOAD_ODD },
    {"epr15617.4" , pcm, 0x040000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver tantr_driver =
{
    "tantr\0",
    "Tant-R (Puzzle & Action) (Japan)\0",
    tantr_rom,
    tantr_init,
};

/*--------------------------------------------------------------------------*/
/* Tant-R (Puzzle & Action) (Japan Bootleg?)                                */
/*--------------------------------------------------------------------------*/

t_rominfo tantrbl_rom[] =
{
    {"pa_e10.bin", rom, 0x080000, 0x000000, ROM_LOAD_EVEN }, /* Unique to Bootleg */
    {"pa_f11.bin", rom, 0x080000, 0x000000, ROM_LOAD_ODD },  /* Unique to Bootleg */
    {"mpr15616.34", rom, 0x080000, 0x100000, ROM_LOAD_EVEN },
    {"mpr15615.33", rom, 0x080000, 0x100000, ROM_LOAD_ODD },
    {"pa_e02.bin", pcm, 0x020000, 0x000000, ROM_LOAD }, /* Unique to Bootleg */
    {"pa_e03.bin", pcm, 0x020000, 0x020000, ROM_LOAD }, /* Unique to Bootleg */
    { NULL }
};

t_driver tantrbl_driver =
{
    "tantrbl\0",
    "Tant-R (Puzzle & Action) (Japan, Bootleg?)\0",
    tantrbl_rom,
    tantrbl_init,
};

/*--------------------------------------------------------------------------*/
/* Ichident-R (Puzzle & Action 2) (Japan)                                   */
/*--------------------------------------------------------------------------*/

t_rominfo ichident_rom[] =
{
    {"epr16886", rom, 0x080000, 0x000000, ROM_LOAD_EVEN },
    {"epr16885", rom, 0x080000, 0x000000, ROM_LOAD_ODD },
    {"epr16888", rom, 0x080000, 0x100000, ROM_LOAD_EVEN },
    {"epr16887", rom, 0x080000, 0x100000, ROM_LOAD_ODD },
    {"epr16884", pcm, 0x080000, 0x000000, ROM_LOAD },
    { NULL }
};

t_driver ichident_driver =
{
    "ichident\0",
    "Ichident-R (Puzzle & Action 2) (Japan)\0",
    ichident_rom,
    ichident_init,
};

/*--------------------------------------------------------------------------*/
/* Bad Dump ? Spaindumps version of Puzzle & Action (c)1994 Sega            */
/*--------------------------------------------------------------------------*/

t_rominfo spainpa_rom[] =
{
    {"pa_2.bin", rom, 0x040000, 0x000000, ROM_LOAD_EVEN },
    {"pa_1.bin", rom, 0x040000, 0x000000, ROM_LOAD_ODD },
    {"pa_4.bin", rom, 0x040000, 0x100000, ROM_LOAD_EVEN },
    {"pa_3.bin", rom, 0x040000, 0x100000, ROM_LOAD_ODD },
    { NULL }
};

t_driver spainpa_driver =
{
    "spainpa\0",
    "SpainDumps version of Puzzle & Action (Bad Dump?)\0",
    spainpa_rom,
    NULL,
};

/*--------------------------------------------------------------------------*/
/* Game list                                                                */
/*--------------------------------------------------------------------------*/

t_driver *driver_list[] =
{
        &bloxeedc_driver, /* Seems Playable */
    &borench_driver, /* Seems Playable */
    &columns_driver,
    &columns2_driver,
	&ichident_driver,
	&potopoto_driver,
    &puyopuyo_driver,
	&puyopuya_driver,
    &puyopuy2_driver, /* Seems Playable */
	&spainpa_driver, /* Bad Dump? */
    &stkclmns_driver,
    &tantr_driver,
	&tantrbl_driver, /* Seems Playable */
	&tforceac_driver,
	&tfacjp_driver,
    &tfacbl_driver, /* Seems Playable */
    &zunkyou_driver,
    NULL,
};

/*--------------------------------------------------------------------------*/
/* Game patching and configuration                                          */
/*--------------------------------------------------------------------------*/

void bloxeedc_init(void)
{
    palmode = 1;
    gab2 = 1;
}

void borench_init(void)
{
	rom[0x004B2]=0x60;
	rom[0x04ABE]=0x60;
	rom[0x0AFB2]=0x60;
	rom[0x06D80]=0x60;
	rom[0x0724E]=0x60;
    palmode = 1;
    gab2 = 1;
}

void columns_init(void)
{
	rom[0x03572]=0x60;
	rom[0x02FBA]=0x60;
	rom[0x02700]=0x60;
	rom[0x02906]=0x60;
	rom[0x01EBA]=0x60;
    palmode = 1;
    gab2 = 1;
}

void columns2_init(void)
{
	rom[0x00E1E]=0x60;
	rom[0x02FAC]=0x60;
	rom[0x03712]=0x60;
	rom[0x00FC8]=0x60;
	rom[0x01FB6]=0x60;
	rom[0x025B2]=0x60;
	rom[0x027CA]=0x60;
    palmode = 1;
    gab2 = 1;
}

void ichident_init(void)
{
	rom[0x000390]=0x60;
    rom[0x000A06]=0x4E; rom[0x000A07]=0x71;
    rom[0x0004B6]=0x60;
    palmode = 0;
    gab = 0;
    gab2 = 0;
}

void potopoto_init(void)
{
	rom[0x03CAA]=0x60;
    palmode = 1;
    gab2 = 1;
    gab = 0;
}

void puyopuyo_init(void)
{
	rom[0x0C1C]=0x60;
	rom[0x010F8]=0x4E;rom[0x010F9]=0x71;rom[0x010FA]=0x4E;rom[0x010FB]=0x71;
    palmode = 1;
    gab2 = 1;
}

void puyopuya_init(void)
{
	rom[0x0C1C]=0x60;
	rom[0x010F8]=0x4E;rom[0x010F9]=0x71;rom[0x010FA]=0x4E;rom[0x010FB]=0x71;
    palmode = 1;
    gab2 = 1;
}

void puyopuy2_init(void)
{
	rom[0x000D78]=0x4E; rom[0x000D79]=0x71;
	rom[0x018886]=0x4E; rom[0x018887]=0x71;
    palmode = 1;
    gab = 0;
    gab2 = 0;
}

void stkclmns_init(void)
{
    rom[0x0A514]=0x4E;
    rom[0x0A515]=0x71;
    rom[0x0A516]=0x4E;
    rom[0x0A517]=0x71;
    rom[0x0A518]=0x4E;
    rom[0x0A519]=0x71;
    rom[0x03456]=0x60;
    rom[0x0058A]=0x60;
    rom[0x00E82]=0x60;
    rom[0x0CA36]=0x60;
    rom[0x147CE]=0x60;
    rom[0x14B86]=0x60;
    rom[0x16810]=0x60;
    palmode = 1;
    gab = 0;
    gab2 = 1;
}

void tantr_init(void)
{
    palmode = 1;
    gab = 0;
    gab2 = 0;
}

void tantrbl_init(void)
{
    palmode = 1;
    gab = 0;
    gab2 = 0;
}

void tforceac_init(void)
{
    palmode = 1;
    gab2 = 1;    
}

void tfacjp_init(void)
{
    palmode = 1;
    gab2 = 1;
}

void tfacbl_init(void)
{
    palmode = 1;
    gab = 0;
    gab2 = 0;
}

void zunkyou_init(void)
{
	rom[0x000AE8]=0x60;
    palmode = 1;
    gab = 0;
    gab2 = 1;
}


/*--------------------------------------------------------------------------*/
/* Driver loading functions                                                 */
/*--------------------------------------------------------------------------*/


/* Print needed files for a given driver */
void print_driver_files(char *name)
{
    int i, j = 0;

    for(i = 0; driver_list[i] != NULL; i += 1)
    {
        if(stricmp(name, driver_list[i]->drvname) == 0)
        {
            printf(" `%s' requires the following files:\n", driver_list[i]->name);
            while(driver_list[i]->rom[j].name != NULL)
            {
                printf("\t%-12s\t%d bytes\n", driver_list[i]->rom[j].name, driver_list[i]->rom[j].size);
                j += 1;
            }
        }
    }
}


void print_driver_list(void)
{
    int i;
    for(i = 0; driver_list[i] != NULL; i += 1)
        printf("%s    \t  `%s'\n", driver_list[i]->drvname, driver_list[i]->name);
}


int load_driver(char *name)
{
    int i;

    /* Initialize driver specific variables */
    dip_used = dip_bank = 0x00;
    dip_switch[0] = dip_switch[1] = 0xFF;
    game_driver = NULL;

    for(i = 0; driver_list[i] != NULL; i += 1)
    {
        if(stricmp(name, driver_list[i]->drvname) == 0)
        {
            char path[0x100];

            /* Load ROM files */
            error("Loading game `%s'\n", driver_list[i]->name);
            game_driver = driver_list[i];
            if(!load_rom(&driver_list[i]->rom[0])) return (0);

            /* Load DIP switch data if it exists */
            sprintf(path, "%s/%s.dip", dirs.dip, game_driver->drvname);
            if(file_exist(path))
            {
                dip_used = 1;
                load_file(path, dip_switch, 2);
            }
        }
    }

    /* Nothing was loaded.. */
    if(game_driver == NULL) return (0);

    return (1);
}


void free_driver(void)
{
    /* Save DIP switch data if modified */
    if(dip_used) {
        char path[0x100];
        sprintf(path, "%s/%s.dip", dirs.dip, game_driver->drvname);
        save_file(path, dip_switch, 2);
        error("saved dip_switch to `%s' for `%s'\n", path, game_driver->name);
    }

    game_driver = NULL;

    memset(rom, 0, ROM_SIZE);
    memset(pcm, 0, PCM_SIZE);
}


void swap(byte *src, byte *dst, int length, int odd)
{
    int count;
    for(count = 0; count < length; count += 1)
        dst[(count << 1) | (odd)] = src[(count)];
}


int load_rom(t_rominfo *p)
{
    int ret = 0, i = 0;
    byte *buf = NULL;
    char file_path[0x100];
    char archive_path[0x100];

    /* See if archive exists */
    sprintf(archive_path, "%s/%s.zip", dirs.zip, game_driver->drvname);
    if(file_exist(archive_path))
    {
        /* Load compressed files from archive */
        while(p[i].name != NULL)
        {
            printf("Loading file `%s' from `%s'\n", p[i].name, archive_path);
            buf = malloc(p[i].size);
            if(!buf) { printf("Couldn't allocate %d bytes for file buffer\n", p[i].size); return (0); };
            ret = load_zip(archive_path, p[i].name, buf, p[i].size);
            if(!ret) { printf("Error loading file `%s' from archive `%s'\n", p[i].name, archive_path); return (0); };
            switch(p[i].flags)
            {
                case ROM_LOAD:
                    memcpy(&p[i].mem[p[i].offset], buf, p[i].size);
                    break;
                case ROM_LOAD_EVEN:
                    swap(buf, &p[i].mem[p[i].offset], p[i].size, 0);
                    break;
                case ROM_LOAD_ODD:
                    swap(buf, &p[i].mem[p[i].offset], p[i].size, 1);
                    break;
            }
            free(buf);
            i += 1;
        }
    }
    else
    {
        /* Load uncompressed files from their respective directories */
        while(p[i].name != NULL)
        {
            sprintf(file_path, "%s/%s/%s", dirs.rom, game_driver->drvname, p[i].name);
            printf("Loading file `%s'\n", file_path);
            buf = malloc(p[i].size);
            if(!buf) { printf("Couldn't allocate %d bytes for file buffer\n", p[i].size); return (0); };
            ret = load_file(file_path, buf, p[i].size);
            if(!ret) { printf("Error loading file`%s'\n", file_path); return (0); };
            switch(p[i].flags)
            {
                case ROM_LOAD:
                    memcpy(&p[i].mem[p[i].offset], buf, p[i].size);
                    break;
                case ROM_LOAD_EVEN:
                    swap(buf, &p[i].mem[p[i].offset], p[i].size, 0);
                    break;
                case ROM_LOAD_ODD:
                    swap(buf, &p[i].mem[p[i].offset], p[i].size, 1);
                    break;
            }
            free(buf);
            i += 1;
        }
    }

    return (1);
}
