
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "upd7759.h"

/* number of samples stuffed into the rom */
static unsigned char numsam;

/* playback rate for the streams interface */
/* BASE_CLOCK or a multiple (if oversampling is active) */
static int emulation_rate;

static int base_rate;
/* define the output rate */
#define CLOCK_DIVIDER	80

/* signal fall off factor */
#define FALL_OFF(n) 	((n)-(((n)+7)/8))

#define SIGNAL_BITS 	15	/* signal range */
#define SIGNAL_MAX		(0x7fff >> (15-SIGNAL_BITS))
#define SIGNAL_MIN		-SIGNAL_MAX

#define STEP_MAX		32
#define STEP_MIN		0

#define DATA_MAX		512

struct UPD7759sample
{
	unsigned int offset;	/* offset in that region */
	unsigned int length;    /* length of the sample */
	unsigned int freq;		/* play back freq of sample */
};


/* struct describing a single playing ADPCM voice */
struct UPD7759voice
{
	int playing;            /* 1 if we are actively playing */
	unsigned char *base;    /* pointer to the base memory location */
	int mask;               /* mask to keep us within the buffer */
	int sample; 			/* current sample number (sample data in slave mode) */
	int freq;				/* current sample playback freq */
	int count;              /* total samples to play */
	int signal;             /* current ADPCM signal */
    int step;               /* current ADPCM step */
	int counter;			/* sample counter */
	void *timer;			/* timer used in slave mode */
	int data[DATA_MAX]; 	/* data array used in slave mode */
	unsigned head;			/* head of data array used in slave mode */
	unsigned tail;			/* tail of data array used in slave mode */
	unsigned available;
};

/* global pointer to the current interface */
UPD7759_interface *upd7759_intf;

/* array of ADPCM voices */
static struct UPD7759voice updadpcm[MAX_UPD7759];

/* stores the current sample number */
static int sampnum[MAX_UPD7759];

/* step size index shift table */
#define INDEX_SHIFT_MAX 16
static int index_shift[INDEX_SHIFT_MAX] = {
	0,	 1,  2,  3,  6,  7, 10, 15,
	0,  15, 10,  7,  6,  3,  2,  1
};

/* lookup table for the precomputed difference */
static int diff_lookup[(STEP_MAX+1)*16];

/*
 *   Compute the difference table
 */

static void ComputeTables (void)
{
	/* nibble to bit map */
    static int nbl2bit[16][4] = {
		{ 1, 0, 0, 0}, { 1, 0, 0, 1}, { 1, 0, 1, 0}, { 1, 0, 1, 1},
		{ 1, 1, 0, 0}, { 1, 1, 0, 1}, { 1, 1, 1, 0}, { 1, 1, 1, 1},
		{-1, 0, 0, 0}, {-1, 0, 0, 1}, {-1, 0, 1, 0}, {-1, 0, 1, 1},
		{-1, 1, 0, 0}, {-1, 1, 0, 1}, {-1, 1, 1, 0}, {-1, 1, 1, 1},
	};
    int step, nib;

	/* loop over all possible steps */
	for (step = 0; step <= STEP_MAX; step++)
	{
        /* compute the step value */
		int stepval = 6 * (step+1) * (step+1);
		/* loop over all nibbles and compute the difference */
		for (nib = 0; nib < 16; nib++)
		{
			diff_lookup[step*16 + nib] = nbl2bit[nib][0] *
				(stepval   * nbl2bit[nib][1] +
				 stepval/2 * nbl2bit[nib][2] +
				 stepval/4 * nbl2bit[nib][3] +
				 stepval/8);
        }
    }
}



static int find_sample(int num, int sample_num,struct UPD7759sample *sample)
{
	int j;
	int nextoff = 0;
	unsigned char *memrom;
	unsigned char *header;   /* upd7759 has a 4 byte what we assume is an identifier (bytes 1-4)*/
	unsigned char *data;


//  memrom = memory_region(upd7759_intf->region[num]);
    memrom = upd7759_intf->rom;

	numsam = (unsigned int)memrom[0]; /* get number of samples from sound rom */
	header = &(memrom[1]);

	if (memcmp (header, "\x5A\xA5\x69\x55",4) == 0)
	{
	}
	else
	{
	}

	/* move the header pointer to the start of the sample offsets */
	header = &(memrom[5]);


	if (sample_num > numsam) return 0;	/* sample out of range */


	nextoff = 2 * sample_num;
	sample->offset = ((((unsigned int)(header[nextoff]))<<8)+(header[nextoff+1]))*2;
//  data = &memory_region(upd7759_intf->region[num])[sample->offset];
    data = &upd7759_intf->rom[sample->offset];
	/* guesswork, probably wrong */
	j = 0;
	if (!data[j]) j++;
	if ((data[j] & 0xf0) != 0x50) j++;

	// Added and Modified by Takahiro Nogi. 1999/10/28
#if 0	// original
	switch (data[j])
	{
		case 0x53: sample->freq = 8000; break;
		case 0x59: sample->freq = 6000; break;
		case 0x5f: sample->freq = 5000; break;
		default:
			sample->freq = 5000;
	}
#else	// modified by Takahiro Nogi. 1999/10/28
	switch (data[j] & 0x1f)
	{
		case 0x13: sample->freq = 8000; break;
		case 0x19: sample->freq = 6000; break;
		case 0x1f: sample->freq = 5000; break;
		default:				// ???
			sample->freq = 5000;
	}
#endif

	if (sample_num == numsam)
	{
		sample->length = 0x20000 - sample->offset;
	}
	else
		sample->length = ((((unsigned int)(header[nextoff+2]))<<8)+(header[nextoff+3]))*2 -
							((((unsigned int)(header[nextoff]))<<8)+(header[nextoff+1]))*2;

	return 1;
}


/*
 *   Start emulation of several ADPCM output streams
 */


#if 1
int UPD7759_sh_start (UPD7759_interface *intf)
{
	int i;

    /* compute the difference tables */
	ComputeTables ();

    /* copy the interface pointer to a global */
	upd7759_intf = intf;
    base_rate = intf->clock_rate / CLOCK_DIVIDER;

	emulation_rate = base_rate;

	memset(updadpcm,0,sizeof(updadpcm));
	for (i = 0; i < intf->num; i++)
	{
		updadpcm[i].mask = 0xffffffff;
		updadpcm[i].signal = 0;
		updadpcm[i].step = 0;
		updadpcm[i].counter = emulation_rate / 2;
	}
	return 0;
}
#endif

/*
 *   Stop emulation of several UPD7759 output streams
 */

void UPD7759_sh_stop (void)
{
}


/*
 *   Update emulation of an uPD7759 output stream
 */
void UPD7759_update (int chip, unsigned short int *buffer, int left)
{
	struct UPD7759voice *voice = &updadpcm[chip];
	int i;

	/* see if there's actually any need to generate samples */

    if (left > 0)
	{
        /* if this voice is active */
		if (voice->playing)
		{
			voice->available -= left;
			if( upd7759_intf->mode == UPD7759_SLAVE_MODE )
			{
			}
			else
			{
				unsigned char *base = voice->base;
                int val;

                while( left > 0 )
				{
					/* compute the new amplitude and update the current voice->step */
					val = base[(voice->sample / 2) & voice->mask] >> (((voice->sample & 1) << 2) ^ 4);
					voice->step = FALL_OFF(voice->step) + index_shift[val & (INDEX_SHIFT_MAX-1)];
					if (voice->step > STEP_MAX) voice->step = STEP_MAX;
					else if (voice->step < STEP_MIN) voice->step = STEP_MIN;
					voice->signal = FALL_OFF(voice->signal) + diff_lookup[voice->step * 16 + (val & 15)];
					if (voice->signal > SIGNAL_MAX) voice->signal = SIGNAL_MAX;
					else if (voice->signal < SIGNAL_MIN) voice->signal = SIGNAL_MIN;

					while (voice->counter > 0 && left > 0)
					{
						*buffer++ = voice->signal;
						voice->counter -= voice->freq;
						left--;
					}

					voice->counter += emulation_rate;

					/* next! */
					if( ++voice->sample > voice->count )
					{
						while (left-- > 0)
						{
							*buffer++ = voice->signal;
							voice->signal = FALL_OFF(voice->signal);
						}
						voice->playing = 0;
						break;
					}
				}
            }
		}
		else
		{
			/* voice is not playing */
			for (i = 0; i < left; i++)
				*buffer++ = voice->signal;
		}
	}
}

/************************************************************
 UPD7759_message_w

 Store the inputs to I0-I7 externally to the uPD7759.

 I0-I7 input the message number of the message to be
 reproduced. The inputs are latched at the rising edge of the
 !ST input. Unused pins should be grounded.

 In slave mode it seems like the ADPCM data is stuffed
 here from an external source (eg. Z80 NMI code).
 *************************************************************/

void UPD7759_message_w (int num, int data)
{
	/* range check the numbers */
	if( num >= upd7759_intf->num )
	{
		return;
	}

	if (upd7759_intf->mode == UPD7759_SLAVE_MODE)
	{
	}
	else
	{
		sampnum[num] = data;
    }
}


/************************************************************
 UPD7759_start_w

 !ST pin:
 Setting the !ST input low while !CS is low will start
 speech reproduction of the message in the speech ROM locations
 addressed by the contents of I0-I7. If the device is in
 standby mode, standby mode will be released.
 NOTE: While !BUSY is low, another !ST will not be accepted.
 *************************************************************/

void UPD7759_start_w (int num, int data)
{
	struct UPD7759voice *voice = updadpcm + num;

	/* range check the numbers */
	if( num >= upd7759_intf->num )
	{
		return;
	}

	/* handle the slave mode */
	if (upd7759_intf->mode == UPD7759_SLAVE_MODE)
	{
	}
	else
	{
		struct UPD7759sample sample;

		/* if !ST is high, do nothing */ /* EHC - 13/08/99 */
		if (data > 0)
			return;

		/* bail if the chip is busy */
		if (voice->playing)
			return;

		/* find a match */
		if (find_sample(num,sampnum[num],&sample))
		{
			/* update the  voice */
			voice->freq = sample.freq;
			/* set up the voice to play this sample */
			voice->playing = 1;
//          voice->base = &memory_region(upd7759_intf->region[num])[sample.offset];
            voice->base = &upd7759_intf->rom[sample.offset];
			voice->sample = 0;
			/* sample length needs to be doubled (counting nibbles) */
			voice->count = sample.length * 2;

			/* also reset the chip parameters */
			voice->step = 0;
			voice->counter = emulation_rate / 2;

			return;
		}

	}
}

/************************************************************
 UPD7759_data_r

 External read data from the UPD7759 memory region based
 on voice->base. Used in slave mode to retrieve data to
 stuff into UPD7759_message_w.
 *************************************************************/

int UPD7759_data_r(int num, int offs)
{
	struct UPD7759voice *voice = updadpcm + num;

    /* range check the numbers */
	if( num >= upd7759_intf->num )
	{
		return 0x00;
    }

	if ( voice->base == NULL )
	{
		return 0x00;
	}

	return voice->base[offs];
}

/* helper functions to be used as memory read handler function pointers */
int UPD7759_0_data_r(int offs) { return UPD7759_data_r(0, offs); }
int UPD7759_1_data_r(int offs) { return UPD7759_data_r(1, offs); }


void UPD7759_reset_w (int num, int data)
{
	struct UPD7759voice *voice = updadpcm + num;

    if( num >= upd7759_intf->num ) return;
    if (data > 0)  return;
	voice->playing = 0;
}
