
#include "shared.h"

/* Global data */
t_bitmap bitmap;
t_input input;
t_snd snd;
t_path dirs;


/* Define how the uPD7759 sample chip is set up */
UPD7759_interface upd7759_interface =
{
    1,                          /* One chip */
    640000*6,                   /* Clock (unsure) */
    { 50 },                     /* Volume */
    pcm,                        /* Memory pointer (gen.h) */
    UPD7759_STANDALONE_MODE,    /* Chip mode */
};

/*--------------------------------------------------------------------------*/

int system_init(int sample_rate)
{
    audio_init(sample_rate);

    genesis_init();
    make_table();
    vdp_reset();
    genesis_reset();

    return (1);
}


int audio_init(int sample_rate)
{
    if(sample_rate)
    {
        snd.sample_rate = sample_rate;
        snd.buffer_size = (snd.sample_rate / 60);

        /* YM3438 sound buffer */
        snd.fm_buffer[0]  = malloc(snd.buffer_size * 2);
        snd.fm_buffer[1]  = malloc(snd.buffer_size * 2);
        if(!snd.fm_buffer[0] || !snd.fm_buffer[1]) return (0);

        /* SN76489 sound buffer */
        snd.psg_buffer = malloc(snd.buffer_size * 2);
        if(!snd.psg_buffer) return (0);

        /* uPD7759 sound buffer */
        snd.pcm_buffer = malloc(snd.buffer_size * 2);
        if(!snd.pcm_buffer) return (0);

        /* Mixed output sound buffer */
        snd.buffer[0]  = malloc(snd.buffer_size * 2);
        snd.buffer[1]  = malloc(snd.buffer_size * 2);
        if(!snd.buffer[0] || !snd.buffer[1]) return (0);

        /* Initialize YM2612 (YM3438) sound */
        YM2612Init(1, FM_CLOCK, snd.sample_rate, 0, 0);

        /* Initialize SN76489 (VDP) sound */
        SN76496_init(0, PSG_CLOCK, 255, snd.sample_rate);

        /* Initialize uPD7759 sound */
        UPD7759_sh_start(&upd7759_interface);

        snd.enabled = 1;
    }

    return (1);
}


void system_reset(void)
{
    genesis_reset();
    vdp_reset();

    /* Reset the YM3438 sound emulation */
    if(snd.enabled)
    {
        YM2612ResetChip(0);
    }
}


void system_shutdown(void)
{
    /* Shut down the virtual console emulation */
    genesis_trash();
 
    /* Free sound data */
    if(snd.enabled)
    {
        YM2612Shutdown();
        if(snd.psg_buffer) free(snd.psg_buffer);
        if(snd.pcm_buffer) free(snd.pcm_buffer);
        if(snd.buffer[0]) free(snd.buffer[0]);
        if(snd.buffer[1]) free(snd.buffer[1]);
        if(snd.fm_buffer[0]) free(snd.fm_buffer[0]);
        if(snd.fm_buffer[1]) free(snd.fm_buffer[1]);
    }
}
